from ._AbstractGidObject import _AbstractGidObject
from .Line import Line
from .Point import Point
import logging
from typing import List
import logging
from typing import List

from .Line import Line
from .Point import Point
from ._AbstractGidObject import _AbstractGidObject

logger = logging.getLogger(__name__)


class Surface(_AbstractGidObject):
    local_id: int = 1
    global_id: int = 1
    surfaces: List = []

    def __init__(self, lines: List[Line]):
        self.local_id = Surface.local_id
        self.lines = lines
        self.id = None
        Surface.local_id += 1

    def add(self):
        self.id = Surface.global_id
        logger.info(f'Adding surface {self.id} connecting {self.lines}')
        Surface.global_id += 1
        Surface.surfaces.append(self)
        export_str = 'Mescape Geometry Create NurbsSurface\n'
        for line in self.lines:
            export_str += f'{line.id} '
        export_str += '\n'
        return export_str

    def extrude(self, start_point: Point, end_point: Point, end_object):
        logger.info(f'Extruding {self} using direction {start_point.coords} -> {end_point.coords}')
        export_str = f'Mescape Utilities Copy Surfaces DoExtrude {end_object} MaintainLayers Translation FNoJoin {start_point.coords_comma} FNoJoin {end_point.coords_comma}\n'
        export_str += f'{self.id}\n'
        extra_surfaces = 5
        extra_lines = 2 * len(self.lines)
        extra_points = len(self.lines)
        Surface.global_id += extra_surfaces
        Line.global_id += extra_lines
        Point.global_id += extra_points
        return export_str

    def __repr__(self):
        if self.id:
            return f"Surface {self.id}"
        else:
            return f"Surface (local) {self.local_id}"




import unittest
from pydelling.readers import StreamlineReader
from pydelling.config import config
from pydelling.utils import test_data_path
from pathlib import Path


class StreamlineReaderCase(unittest.TestCase):
    def setUp(self) -> None:
        self.stream_line_reader = StreamlineReader(filename=test_data_path() / config.streamline_reader.file)

    def test_read_case(self):
        self.assertEqual(self.stream_line_reader.raw_data["Points:0"].loc[0], 0.066709)

    def test_generate_streams(self):
        self.assertEqual(self.stream_line_reader.stream_data.get_group(0).iloc[1, 0], 1.0)

    def test_compute_arrival_times(self):
        arrival_times = self.stream_line_reader.compute_arrival_times()
        self.assertEqual(arrival_times[10], 14114.0)

    def test_write_csv(self):
        write_path = Path.cwd() / "test_streamlines.csv"
        self.stream_line_reader.dump_to_csv(write_path)
        write_path.unlink()



if __name__ == '__main__':
    unittest.main()



import numpy as np
from ..preprocessing.mesh_preprocessor import MeshPreprocessor
import logging
import logging

import numpy as np

from ..preprocessing.mesh_preprocessor import MeshPreprocessor

logger = logging.getLogger(__name__)
from tqdm import tqdm
from pathlib import Path
import streamlit as st


class SmeshReader(MeshPreprocessor):
    has_kd_tree = False

    def __init__(self, filename, kd_tree=True, st_file=False):
        super().__init__()
        self.is_streamlit = st_file

        if Path(filename).suffix == '.smesh':
            self.read_file(filename)
            if kd_tree:
                self.create_kd_tree()
                self.has_kd_tree = True
        else:
            self.load(filename)

    def read_file(self, filename):
        with open(filename, 'r') as f:
            first_line = f.readline()
            number_of_nodes = int(first_line.split()[0])
            if self.is_streamlit:
                empty_progress = st.empty()
                empty_progress.progress(0)
                st.write(f'Reading {number_of_nodes} nodes')
            for _ in tqdm(range(number_of_nodes), desc='Reading nodes'):
                node_info = f.readline().split()
                prop = int(np.ceil(_ / number_of_nodes * 100))
                if prop > 100:
                    prop = 100
                if self.is_streamlit:
                    if prop % 10 == 0:
                        empty_progress.progress(prop)
                self.aux_nodes[int(node_info[0])] = np.array([float(node_info[1]), float(node_info[2]), float(node_info[3])])
            # Read elements
            number_of_elements = int(f.readline().split()[0])
            if self.is_streamlit:
                empty_progress_elements = st.empty()
                empty_progress_elements.progress(0)
                st.write(f'Generating {number_of_elements} elements')
            for _ in tqdm(range(number_of_elements), desc='Reading elements'):
                prop = int(np.ceil(_ / number_of_elements * 100))
                if prop > 100:
                    prop = 100
                if self.is_streamlit:
                    if prop % 5 == 0:
                        empty_progress_elements.progress(prop)
                element_info = f.readline().split()
                element_type = int(element_info[0])
                element_node_ids = np.array([int(node_id) for node_id in element_info[1:element_type + 1]])
                element_coords = [self.aux_nodes[node_id] for node_id in element_node_ids]
                element_node_ids -= 1  # Node IDs should start at 0
                if element_type == 4:
                    self.add_tetrahedra(node_ids=element_node_ids, node_coords=element_coords)
                elif element_type == 5:
                    self.add_pyramid(node_ids=element_node_ids, node_coords=element_coords)
                elif element_type == 6:
                    self.add_wedge(node_ids=element_node_ids, node_coords=element_coords)
                elif element_type == 8:
                    self.add_hexahedra(node_ids=element_node_ids, node_coords=element_coords)



"""
Base interface for a reader class
"""
import logging

import numpy as np

from pydelling.readers import BaseReader

logger = logging.getLogger(__name__)
from pydelling.config import config
import logging
from pathlib import Path
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
logger = logging.getLogger(__name__)


class PflotranObservationPointReader(BaseReader):
    observation_point: np.ndarray
    observation_boundary: str
    observation_node: int
    variables: dict
    def __init__(self, filename=None):
        self.filename = Path(filename) if filename else Path(config.pflotran_reader.filename)
        logger.info(f"Reading PFLOTRAN observation point results file from {self.filename}")
        super().__init__(filename=self.filename)
        self.results = self.data.copy()

        # self.results = {}
        # for time in self.time_keys:
        #     self.results[time] = PflotranResults(time=time, data=self.data[self.time_keys[time]])
        # self.variables = list(self.results[self.time_values[0]].variable_keys)

    def open_file(self, filename):
        self.data: pd.DataFrame = pd.read_csv(self.filename,
                                              skiprows=1,
                                              header=None,
                                              delim_whitespace=True
                                              )
        header = pd.read_csv(self.filename,
                             nrows=0
                             ).columns.tolist()
        self.data.columns = header

        # Rename time column
        time_column: str = self.data.columns.to_list()[0]
        # self.data.rename(columns={time_column: time_column.strip().strip('"')}, inplace=True)
        # # Get observation point information
        # temp_column_name: str = self.data.columns[1]
        # temp_column_name = temp_column_name.replace('(', '').replace(')', '')
        # splitted_column_name = temp_column_name.split(' ')
        # self.observation_point = np.array([splitted_column_name[-3],
        #                                   splitted_column_name[-2],
        #                                   splitted_column_name[-3]],
        #                                   ).astype(float)
        # self.observation_node = int(splitted_column_name[-4])
        # self.observation_boundary = splitted_column_name[-5]
        # Get output column names
        # self.variables = {}
        # for column in self.data.columns:
        #     if 'Time' in column:
        #         continue
        #     test_column_name: str = column
        #     test_column_name_split = test_column_name.split('-')[1].split()
        #     variable_str = []
        #     flag_list = ['east', 'west', 'north', 'south', 'top', 'bottom']
        #     for piece in test_column_name_split:
        #         if piece in flag_list:
        #             break
        #         else:
        #             variable_str.append(piece)
        #     variable_str = ' '.join(variable_str)
        #     self.variables[variable_str] = column
        #     self.data.rename(columns={column: variable_str}, inplace=True)

    @property
    def mineral_names(self):
        temp_keys = [key for key in self.variables if 'VF' in key]
        return temp_keys

    @property
    def total_species_names(self):
        temp_keys = [key for key in self.variables if 'Total' in key]
        return temp_keys

    @property
    def free_species_names(self):
        temp_keys = [key for key in self.variables if 'Free' in key]
        return temp_keys

    def plot_variable(self, variable) -> plt.Axes:
        logger.info(f'Creating lineplot of {variable}')
        plt.clf()
        lineplot: plt.Axes = sns.lineplot(x=self.time_series,
                                y=self.results[variable])
        lineplot.set_label(f'{variable}')
        return lineplot

    def to_csv(self, filename='postprocess/results.csv', variables=None) -> pd.DataFrame:
        self.create_postprocess_dict()
        logger.info(f'Exporting results to csv')
        if variables:
            plot_results: pd.DataFrame = self.results[variables]
            plot_results.to_csv(filename)
        else:
            self.results.to_csv(filename, index=False)

        # print(self.variables)

    def get_mineral_vf_key(self, mineral):
        """
        Returns the correct key of the mineral volume fraction name
        Args:
            mineral: mineral name

        Returns:
            mineral volume fraction key
        """
        return f"{mineral}_VF [m^3 mnrl_m^3 bulk]"

    def get_mineral_rate_key(self, mineral):
        """
        Returns the correct key of the mineral rate name
        Args:
            mineral: mineral name

        Returns:
            mineral rate key
        """
        return f"{mineral}_Rate [mol_m^3_sec]"

    def get_mineral_si_key(self, mineral):
        """
        Returns the correct key of the mineral si name
        Args:
            mineral: mineral name

        Returns:
            mineral si key
        """
        return f"{mineral}_SI"

    def get_primary_species_key(self, species):
        """
        Returns the correct key of the given species name
        Args:
            species: specie name

        Returns:
            specie key
        """
        return f"Total_{species}"

    @property
    def time_series(self):
        return self.results.iloc[:, 0]



from __future__ import annotations

import functools
import logging

from .Line import Line
from .Point import Point
from .Surface import Surface
from ._AbstractGidObject import _AbstractGidObject

logger = logging.getLogger(__name__)
from typing import Union, List

class GidObject(object):
    """
    This class can be inherited to create new GiD objects
    It stores different GiD objects and generates batch files to generate the geometry automatically on GiD
    """
    def __init__(self, *args, **kwargs):
        self.batch_commands = ''
        self.points = []
        self.lines = []
        self.surfaces = []
        self.volumes = []
        self.n_points = 1
        self.n_lines = 1
        self.n_surfaces = 1
        self.n_volumes = 1
        self.pipeline = []
        # self.set_up(*args, **kwargs)

    def construct(self, *args, **kwargs):
        pass

    def set_up(self, *args, **kwargs):
        """
        This method is expected to be changed by the user and it is called when the object is initialized
        """

    def add(self, objects: Union[List[_AbstractGidObject], _AbstractGidObject, GidObject]):
        """
        Adds an object of type AbstractGidObject to the current GidObject instance depending on its type
        """
        if not isinstance(objects, list):
            objects = [objects]

        for obj in objects:
            assert isinstance(obj, _AbstractGidObject) or isinstance(obj, GidObject), 'The given object must be a Point, Line, Surface, Volume or another GidObject'
            if isinstance(obj, Point):
                self.pipeline.append({'add': obj})

            if isinstance(obj, Line):
                self.pipeline.append({'add': obj})

            if isinstance(obj, Surface):
                self.pipeline.append({'add': obj})

            if isinstance(obj, GidObject):
                self.import_gidobject(obj)

    def extrude(self, obj: Union[_AbstractGidObject, GidObject], start_point: Point, end_point: Point, end_object='Volumes'):
        assert isinstance(obj, Surface), 'Only surfaces can be extruded for now'
        self.pipeline.append({'extrude': obj, 'kwargs': {'start_point': start_point, 'end_point': end_point, 'end_object': end_object}})

    def join(self, point_1: Point, point_2: Point):
        join_line: Line = Line(point_1, point_2)
        self.add(join_line)

    def import_gidobject(self, gidobject):
        gidobject.construct()
        self.pipeline += gidobject.pipeline

    def run(self, filename='gid_batch.bch', add_escape=False, internal=False):
        """
        Processes the construct method and generates the bash file
        """
        logger.info('Processing GidObject by runnning the "construct()" method')
        self.construct()  # Generates the pipeline
        for step in self.pipeline:
            keys = list(step.keys())
            step_type = keys[0]
            if step.get('internal'):
                method = getattr(self, step_type)
            else:
                method = getattr(step[step_type], step_type)
            if add_escape:
                method = self.add_escape(method)
            self.batch_commands += method(*step.get('args', {}), **step.get('kwargs', {}))

        self.batch_commands += 'escape'
        if internal:
            return self.batch_commands
        else:
            with open(filename, 'w') as write_file:
                write_file.write(self.batch_commands)


    @staticmethod
    def add_escape(func):
        @functools.wraps(func)
        def decorator(*args, **kwargs):
            func_return = func(*args, **kwargs)
            func_return += 'escape\n'
            return func_return
        return decorator



from typing import List

import numpy as np
import shapely.geometry as geom

from pydelling.utils.geometry import Plane, Segment, Point, Line


class Fracture(object):
    local_id = 0
    eps = 1e-3
    _transmissivity = None
    _storativity = None
    _side_points = None
    _unit_normal_vector = None

    def __init__(self,
                 x=None,
                 y=None,
                 z=None,
                 dip=None,
                 dip_dir=None,
                 size=None,
                 aperture=None,
                 aperture_constant=None,
                 transmissivity_constant=None,
                 storativity_constant=None,
                 normal_vector=None,
                 polygon: List[np.ndarray]=None,
                 ):
        if normal_vector is not None:
            self._unit_normal_vector: np.ndarray = normal_vector
        self.side_points = None
        if dip is not None:
            assert dip_dir is not None
        if size is None:
            assert aperture is not None
        if aperture is None:
            assert size is not None
        # Allow definition of a Fracture object based on a polygon
        if polygon is not None:
            # Assert x, y, z are not provided
            self._side_points = polygon
            centroid = np.mean(polygon, axis=0)
            x = centroid[0]
            y = centroid[1]
            z = centroid[2]
        self.polygon_points = polygon


        self.dip = dip
        self.dip_dir = dip_dir
        self.x_centroid = x
        self.y_centroid = y
        self.z_centroid = z
        self.size = size
        self._aperture = aperture
        self.intersection_dictionary = {}
        self.aperture_constant = aperture_constant
        self.transmissivity_constant = transmissivity_constant
        self.storativity_constant = storativity_constant
        self.aperture = self.compute_aperture()
        self.local_id = Fracture.local_id
        self.side_points = self.get_side_points()
        self.n_side_points = len(self.side_points)


        Fracture.local_id += 1

    def get_side_points_v1(self):
        """
        Finds the side points of the fracture
        Returns: Coordinates of the plane side points
        """
        phi = self.dip_dir / 360 * (2 * np.pi)
        theta = self.dip / 360 * (2 * np.pi)

        u = np.array([np.cos(theta) * np.sin(phi), np.cos(theta) * np.cos(phi), -np.sin(theta)])
        v = np.array([np.cos(phi), - np.sin(phi), 0])

        A = self.centroid + self.size / 2 * (u + v)
        B = self.centroid + self.size / 2 * (u - v)
        C = self.centroid - self.size / 2 * (u + v)
        D = self.centroid - self.size / 2 * (u - v)

        self.side_points = 4

        return np.array([A, B, C, D])

    def get_side_points_v3(self):
        """
        alpha = 45; %strike
        delta = 30; %dip
        w = 4; %width (dipping side)
        L = 9; %Length (side parallel to the surface)

        A = [0,0,0]; % Pivot point at the surface

        H = w*sind(alpha+90);
        V = w*cosd(alpha+90);
        Z = -w*sind(delta);

        B = [L*sind(alpha) L*cosd(alpha) 0]+A;
        C = [L*cosd(90-alpha)+H L*sind(90-alpha)+V Z]+A;
        D = [w*sind(alpha+90) w*cosd(alpha+90) -w*sind(delta)]+A;

        P = [A; B; C; D];
        """
        alpha = self.dip_dir / 360 * (2 * np.pi)
        delta = self.dip / 360 * (2 * np.pi)
        w = self.size
        L = self.size

        A = np.array(self.centroid)
        A[0] = A[0] - np.sin(alpha) * L / 2
        A[1] = A[1] + np.cos(alpha) * L / 2
        A[2] = A[2] + np.sin(delta) * L / 2

        H = w * np.sin(alpha + np.pi / 2)
        V = w * np.cos(alpha + np.pi / 2)
        Z = -w * np.sin(delta)

        B = np.array([L * np.sin(alpha), L * np.cos(alpha), 0]) + A
        C = np.array([L * np.cos(np.pi / 2 - alpha) + H, L * np.sin(np.pi / 2 - alpha) + V, Z]) + A
        D = np.array([w * np.sin(alpha + np.pi / 2), w * np.cos(alpha + np.pi / 2), -w * np.sin(delta)]) + A

        P = np.array([A, B, C, D])

        return P

    def get_side_points_v2(self):
        """
        Finds the side points of the fracture
        Returns: Coordinates of the plane side points
        """
        alpha = self.dip / 360 * (2 * np.pi)
        beta = self.dip_dir / 360 * (2 * np.pi)

        A = self.centroid + np.array([
            + self.size / 2 * (-np.cos(beta) - np.sin(beta) * np.cos(alpha)),
            + self.size / 2 * (np.sin(beta) - np.cos(beta) * np.cos(alpha)),
            self.size / 2 * np.sin(alpha)
        ])

        B = self.centroid + np.array([
            + self.size / 2 * (-np.cos(beta) + np.sin(beta) * np.cos(alpha)),
            + self.size / 2 * (np.sin(beta) + np.cos(beta) * np.cos(alpha)),
            - self.size / 2 * np.sin(alpha)
        ])

        C = self.centroid + np.array([
            + self.size / 2 * (np.cos(beta) + np.sin(beta) * np.cos(alpha)),
            + self.size / 2 * (-np.sin(beta) + np.cos(beta) * np.cos(alpha)),
            - self.size / 2 * np.sin(alpha)
        ])

        D = self.centroid + np.array([
            + self.size / 2 * (np.cos(beta) - np.sin(beta) * np.cos(alpha)),
            + self.size / 2 * (-np.sin(beta) - np.cos(beta) * np.cos(alpha)),
            + self.size / 2 * np.sin(alpha)
        ])

        self.side_points = 4

        return np.array([A, B, C, D])

    def get_side_points(self, method='v1'):
        if self._side_points is not None:
            return self._side_points
        if method == 'v1':
            return self.get_side_points_v1()
        elif method == 'v2':
            return self.get_side_points_v2()
        elif method == 'v3':
            return self.get_side_points_v3()


    def to_obj(self, global_id=0, method='v1'):
        """Converts the fracture to an obj file"""
        side_points = self.get_side_points(method=method)
        if isinstance(side_points[0], np.ndarray):
            side_points = [side_point.tolist() for side_point in side_points]
        obj_string = ''
        for i in range(len(side_points)):
            obj_string += 'v ' + str(side_points[i][0]) + ' ' + str(side_points[i][1]) + ' ' + str(
                side_points[i][2]) + '\n'
        obj_string += f'f '
        for i in range(len(side_points)):
            obj_string += str(global_id + i) + ' '
        obj_string += '\n'
        return obj_string

    @property
    def unit_normal_vector(self):
        """Returns the normal vector of the fracture"""
        if self._unit_normal_vector is None:
            get_side_points = self.get_side_points()
            v1 = get_side_points[1] - get_side_points[0]
            v2 = get_side_points[2] - get_side_points[0]
            cross = np.cross(v1, v2)
            self._unit_normal_vector = cross / np.linalg.norm(cross)
        return self._unit_normal_vector

    def distance_to_point(self, point: np.ndarray):
        """Returns the minimum distance to a point from the fracture plane"""
        a = self.unit_normal_vector[0]
        b = self.unit_normal_vector[1]
        c = self.unit_normal_vector[2]
        d = -np.dot(self.unit_normal_vector, self.centroid)
        return abs(a * point[0] + b * point[1] + c * point[2] + d) / np.sqrt(a ** 2 + b ** 2 + c ** 2)

    def get_bounding_box(self):
        """Returns the bounding box of the fracture"""
        side_points = self.get_side_points()
        x_min = np.min(side_points[:, 0])
        x_max = np.max(side_points[:, 0])
        y_min = np.min(side_points[:, 1])
        y_max = np.max(side_points[:, 1])
        z_min = np.min(side_points[:, 2])
        z_max = np.max(side_points[:, 2])
        return np.array([x_min, x_max, y_min, y_max, z_min, z_max])

    def point_inside_bounding_box(self, point: np.ndarray, scale_factor=0.0):
        """Returns if a point is inside the bounding box of the fracture"""
        bounding_box = self.get_bounding_box()
        lx = bounding_box[1] - bounding_box[0]
        ly = bounding_box[3] - bounding_box[2]
        lz = bounding_box[5] - bounding_box[4]

        if point[0] < bounding_box[0] - scale_factor * lx or point[0] > bounding_box[1] + scale_factor * lx:
            return False
        if point[1] < bounding_box[2] - scale_factor * ly or point[1] > bounding_box[3] + scale_factor * ly:
            return False
        if point[2] < bounding_box[4] - scale_factor * lz or point[2] > bounding_box[5] + scale_factor * lz:
            return False
        else:
            return True

    @property
    def centroid(self):
        return np.array([self.x_centroid, self.y_centroid, self.z_centroid])

    @property
    def polygon(self):
        side_points = self.get_side_points()
        self._polygon = geom.Polygon(side_points)
        return self._polygon


    @property
    def plane(self):
        """Returns the plane of the fracture"""
        if not hasattr(self, '_plane'):
            self._plane = Plane(self.centroid, normal=self.unit_normal_vector)

        return self._plane

    @property
    def corners(self) -> List[Point]:
        """Returns the corners of the fracture"""
        if not hasattr(self, '_corners'):
            self._corners = [Point(point) for point in self.get_side_points()]
        return self._corners

    @property
    def corner_segments(self):
        """Returns the corner lines of the fracture"""
        corner_segments = [
            Segment(self.corners[0], self.corners[1]),
            Segment(self.corners[1], self.corners[2]),
            Segment(self.corners[2], self.corners[3]),
            Segment(self.corners[3], self.corners[0])
        ]
        return corner_segments

    @property
    def corner_lines(self):
        """Returns the corner lines of the fracture"""
        corner_segments = [
            Line(self.corners[0], self.corners[1]),
            Line(self.corners[1], self.corners[2]),
            Line(self.corners[2], self.corners[3]),
            Line(self.corners[3], self.corners[0]),
        ]
        return corner_segments

    def contains(self, point: Point):
        """Returns if a point is inside the fracture"""
        q1, q2, q3, q4 = self.corners
        q1: Point

        largest_normal_index = self.largest_index_normal_vector
        q1_hat = np.delete(q1, largest_normal_index)
        q2_hat = np.delete(q2, largest_normal_index)
        q3_hat = np.delete(q3, largest_normal_index)
        q4_hat = np.delete(q4, largest_normal_index)
        p_hat = np.delete(point, largest_normal_index)
        u0 = p_hat[0]
        u1 = q1_hat[0]
        u2 = q2_hat[0]
        u3 = q3_hat[0]
        u4 = q4_hat[0]
        v0 = p_hat[1]
        v1 = q1_hat[1]
        v2 = q2_hat[1]
        v3 = q3_hat[1]
        v4 = q4_hat[1]

        s1 = (v1 - v2) * u0 + (u2 - u1) * v0 + v2 * u1 - u2 * v1
        s2 = (v2 - v3) * u0 + (u3 - u2) * v0 + v3 * u2 - u3 * v2
        s3 = (v3 - v4) * u0 + (u4 - u3) * v0 + v4 * u3 - u4 * v3
        s4 = (v4 - v1) * u0 + (u1 - u4) * v0 + v1 * u4 - u1 * v4

        s = np.array([s1, s2, s3, s4])

        equal_sign = np.all(s >= -self.eps) if s[0] >= -self.eps else np.all(s <= self.eps)

        if equal_sign:
            return True
        else:
            return False

    def shift(self, x, y, z):
        """Shifts the fracture"""
        self.x_centroid += x
        self.y_centroid += y
        self.z_centroid += z
        if self._side_points is not None:
            self._side_points = [point + np.array([x, y, z]) for point in self._side_points]

    @property
    def largest_index_normal_vector(self):
        """Returns the largest coordinate index of the normal vector"""
        return np.argmax(self.unit_normal_vector)

    def compute_aperture(self, const=3.020E-3):
        """Computes the aperture of the fracture"""
        if self._aperture is not None:
            return self._aperture
        else:
            # const = config.globals.constants
            # computed_aperture = np.power((12 * const.mu
            #                              * config.globals.constitutive_laws.transmissivity.a
            #                              * np.log10(self.size / 2.0) ** 2) / (const.rho * const.g), 1/3)
            computed_aperture = self.aperture_constant * np.log10(self.size / 2.0)
            return computed_aperture

    @property
    def transmissivity(self):
        """Returns the transmissivity of the fracture"""
        if self._transmissivity is not None:
            return self._transmissivity
        elif self.transmissivity_constant is not None:
            computed_transmissivity = self.transmissivity_constant * (np.log10(self.size / 2.0)) ** 2
            return computed_transmissivity
        else:
            rho = 1000
            g = 9.8
            mu = 8.9E-4
            return (np.power(self.aperture, 2) * rho * g) / (12 * mu)

    @property
    def storativity(self):
        """Returns the storativity of the fracture"""
        if self._storativity is not None:
            return self._storativity
        elif self.storativity_constant is not None:
            computed_storativity = self.storativity_constant * np.log10(self.size / 2.0)
            return computed_storativity
        else:
            return 0.0

    def get_json(self):
        cur_dict = {
            'x': self.x_centroid,
            'y': self.y_centroid,
            'z': self.z_centroid,
            'size': self.size,
            'aperture': self.aperture,
            'dip': self.dip,
            'dip_dir': self.dip_dir,
            'aperture_constant': self.aperture_constant,
            'transmissivity_constant': self.transmissivity_constant,
            'storativity_constant': self.storativity_constant,
            'normal_vector': self._unit_normal_vector.tolist() if self._unit_normal_vector is not None else None,
            'polygon': [point.tolist() for point in self.polygon_points] if self.polygon_points is not None else None,
        }
        return cur_dict







import numpy as np
import pandas as pd

try:
    from paraview.vtk.numpy_interface import dataset_adapter as dsa
    from paraview.vtk.numpy_interface import algorithms as algs
    from paraview import servermanager as sm
    from paraview.simple import *
except:
    pass
from vtk.util.numpy_support import vtk_to_numpy
from typing import List
import logging

logger = logging.getLogger(__name__)


class BaseFilter:
    """
    Class used as a basecase class for the different Paraview filters
    """
    filter_type: str = "VTK_reader"
    counter: int = 0
    filter: object
    vector_keys: List = ["x", "y", "z"]
    x_min: float
    x_max: float
    y_min: float
    y_max: float
    z_min: float
    z_max: float

    def __init__(self, name):
        self.name = name


    @property
    def cell_keys(self):
        vtk_object = sm.Fetch(self.filter)
        vtk_object = dsa.WrapDataObject(vtk_object)
        return vtk_object.CellData.keys()

    @property
    def point_keys(self):
        vtk_object = sm.Fetch(self.filter)
        vtk_object = dsa.WrapDataObject(vtk_object)
        return vtk_object.PointData.keys()

    @property
    def field_keys(self):
        vtk_object = sm.Fetch(self.filter)
        vtk_object = dsa.WrapDataObject(vtk_object)
        return vtk_object.FieldData.keys()

    @property
    def cell_data(self) -> pd.DataFrame:
        vtk_object = sm.Fetch(self.filter)
        vtk_object = dsa.WrapDataObject(vtk_object)
        pd_df = pd.DataFrame()
        for key in self.cell_keys:
            temp_dataset = np.array(vtk_object.CellData[key]).transpose()
            if len(temp_dataset.shape) != 1:
                # The dataset is a vector:
                for idx, vector_element in enumerate(temp_dataset):
                    new_key = f"{key}{self.vector_keys[idx]}"
                    pd_df[new_key] = vector_element
            else:
                pd_df[key] = temp_dataset
        return pd_df

    @property
    def point_data(self) -> pd.DataFrame:
        vtk_object = sm.Fetch(self.filter)
        vtk_object = dsa.WrapDataObject(vtk_object)
        pd_df = pd.DataFrame()
        for key in self.point_keys:
            temp_dataset = np.array(vtk_object.PointData[key]).transpose()
            if len(temp_dataset.shape) != 1:
                # The dataset is a vector:
                for idx, vector_element in enumerate(temp_dataset):
                    new_key = f"{key}{self.vector_keys[idx]}"
                    pd_df[new_key] = vector_element
            else:
                pd_df[key] = temp_dataset
        return pd_df


    @property
    def field_data(self) -> pd.DataFrame:
        vtk_object = sm.Fetch(self.filter)
        vtk_object = dsa.WrapDataObject(vtk_object)
        pd_df = pd.DataFrame()
        for key in self.field_keys:
            temp_dataset = np.array(vtk_object.FieldData[key]).transpose()
            if len(temp_dataset.shape) != 1:
                # The dataset is a vector:
                for idx, vector_element in enumerate(temp_dataset):
                    new_key = f"{key}{self.vector_keys[idx]}"
                    pd_df[new_key] = vector_element
            else:
                pd_df[key] = temp_dataset
        return pd_df

    @property
    def mesh_points(self) -> pd.DataFrame:
        vtk_object = sm.Fetch(self.filter)
        return pd.DataFrame(vtk_to_numpy(vtk_object.GetPoints().GetData()), columns=["x", "y", "z"])

    def add_attribute(self, name: str, value: object):
        """
        This method sets a given attribute to the filter
        Args:
            name: String containing the name of the attribute
            value: Value to set in the attribute
        """

        setattr(self.filter, name, value)
        logger.info(f"Attribute {name} = {value} has been added to {self.name}")

    def to_csv(self, filename=None):
        """
        This method exports the current filter's data into a csv file
        """
        filename = filename if filename else f"{self.name}_filter_data.csv"
        SaveData(filename=filename, proxy=self.filter)
        logger.info(f"Saving {self.name} data into {filename}")



import streamlit as st

from WebAppRunner import WebAppRunner


class TestWebApp(WebAppRunner):
    def construct(self):
        st.header("Test WebApp")


if __name__ == '__main__':
    test_webapp = TestWebApp()
    test_webapp.run()



from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass


class CellDataToPointDataFilter(BaseFilter):
    """
    This class implements the Paraview Integrate Variables filter
    """
    filter_type: str = "Cell_data_to_point_data"
    counter: int = 0

    def __init__(self, input_filter, name):
        super().__init__(name=name)
        CellDataToPointDataFilter.counter += 1
        self.filter = CellDatatoPointData(Input=input_filter)





from abc import ABC, abstractmethod

from pydelling.webapps import WebAppRunner


class BaseComponent(ABC):
    _value = None
    def __init__(self, webapp: WebAppRunner=None, *args, **kwargs):
        self.type = self.__class__.__name__
        self.finished = False
        if webapp:
            self.webapp = webapp
        self.run(*args, **kwargs)

    @abstractmethod
    def run(self, *args, **kwargs):
        """Runs the component"""

    @property
    def value(self):
        """Returns the value of the component"""
        return self._value



from .BaseFilter import BaseFilter
from .AppendArcLengthFilter import AppendArcLengthFilter
from .CalculatorFilter import CalculatorFilter
from .CellDataToPointDataFilter import CellDataToPointDataFilter
from .ClipFilter import ClipFilter
from .CsvReaderFilter import CsvReaderFilter
from .IntegrateVariablesFilter import IntegrateVariablesFilter
from .PlotOverLineFilter import PlotOverLineFilter
from .SaveDataFilter import SaveDataFilter
from .StreamTracerWithCustomSourceFilter import StreamTracerWithCustomSourceFilter
from .TableToPointsFilter import TableToPointsFilter
from .TecplotFilter import TecplotFilter
from .VtkFilter import VtkFilter
from .XDMFFilter import XDMFFilter



from pydelling.readers.iGPReader.utils.geometry_utils import *
from pydelling.utils.geometry import Plane
from .BaseAbstractMeshObject import BaseAbstractMeshObject


class BaseFace(BaseAbstractMeshObject):
    __slots__ = ['node_ids', 'node_coords']
    def __init__(self, node_ids, node_coords):
        self.nodes = np.array(node_ids)
        self.coords = np.array(node_coords)
        self.n_coords = len(node_coords)

    @property
    def area(self):
        return self.compute_area()

    @property
    def centroid(self):
        return self.compute_centroid()

    def compute_area(self):
        """
        This function computes the area of a 3D planar polygon
        :return: Area of a 3D planar polygon
        """
        assert len(self.coords >= 3), "Incorrect number of points, more are needed to form a polygon"
        # Compute normal
        vn = normal_vector(self.coords)
        temp_area_v = np.zeros(shape=3)
        projected_area = 0.0
        for id, point in enumerate(self.coords):
            # Set-up variables
            v1 = self.coords[id % self.n_coords]  # Pv1, assuming P=(0,0,0)
            v2 = self.coords[(id + 1) % self.n_coords]  # Pv2, assuming P=(0,0,0)
            # Compute area
            id_area_v = np.cross(v1, v2)
            projected_area_id = np.dot(vn, id_area_v) / 2.0  # area of small triangle of the face
            projected_area += projected_area_id
        return projected_area

    def compute_centroid(self):
        return np.mean(self.coords, axis=0)

    def compute_centroid_mean(self):
        return np.mean(self.coords, axis=0)

    def plot_face(self):
        from mpl_toolkits.mplot3d import Axes3D
        from mpl_toolkits.mplot3d.art3d import Poly3DCollection
        import matplotlib.pyplot as plt
        x = []
        y = []
        z = []
        # points = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 0.0, 1.0]]
        for point in self.coords:
            x.append(point[0])
            y.append(point[1])
            z.append(point[2])
        fig = plt.figure()
        ax = Axes3D(fig)
        verts = [list(zip(x, y, z))]
        c_mean = np.mean(self.coords, axis=0)
        ax.scatter3D(xs=[self.centroid[0]], ys=[self.centroid[1]], zs=[self.centroid[2]], c="r")
        ax.scatter3D(xs=[c_mean[0]], ys=[c_mean[1]], zs=[c_mean[2]], c="g")
        ax.add_collection3d(Poly3DCollection(verts, alpha=0.5))
        ax.axes.set_xlim3d(np.min(self.coords[:, 0]), np.max(self.coords[:, 0]))
        ax.axes.set_ylim3d(np.min(self.coords[:, 1]), np.max(self.coords[:, 1]))
        ax.axes.set_zlim3d(np.min(self.coords[:, 2]), np.max(self.coords[:, 2]))
        plt.show()

    def intersect_with_plane(self, plane: Plane):
        '''Returns the intersection of the face with the plane'''
        return self.plane.intersect(plane)


    @property
    def unit_normal_vector(self):
        if not hasattr(self, '_unit_normal_vector'):
            v1 = self.coords[1] - self.coords[0]
            v2 = self.coords[2] - self.coords[0]
            self._unit_normal_vector = np.cross(v1, v2) / np.linalg.norm(np.cross(v1, v2))

        return self._unit_normal_vector

    @property
    def edges(self):
        '''Returns the edges of the face'''
        return NotImplementedError('This method is not implemented yet')

    @property
    def edge_vectors(self):
        '''Returns the edge vectors of the face'''
        return NotImplementedError('This method is not implemented yet')


    @property
    def plane(self):
        '''Returns the plane of the face'''
        return Plane(self.centroid, normal=self.unit_normal_vector)



import unittest
import numpy.testing as nptest
import numpy as np
from pydelling.utils.geometry import Line, Plane, Vector


class PlaneCase(unittest.TestCase):
    def test_init(self):
        p = Plane(point=[0, 0, 0], normal=Vector([0, 0, 1]))

    def test_intersection(self):
        plane_1 = Plane(point=[0, 0, 0], normal=Vector([0, 0, 1]))
        plane_2 = Plane(point=[0, 0, 0], normal=Vector([0, 1, 0]))
        intersected_line = plane_1.intersect(plane_2)
        nptest.assert_array_equal(intersected_line.direction_vector, [-1, 0, 0])




if __name__ == '__main__':
    unittest.main()


import unittest
from pydelling.utils.geometry_utils import *
from pydelling.utils.geometry import Point


class GeometryUtilsCase(unittest.TestCase):
    def test_compute_area(self):
        # Test 1
        points = [Point([0.0, 0.0, 0.0]),
                  Point([2.0, 0.0, 0.0]),
                  Point([2.0, 1.0, 0.0]),
                  Point([0.0, 1.0, 0.0])]
        self.assertEqual(compute_polygon_area(points), 2.0)

        # Test 2
        points = [Point([0.0, 0.0, 0.0]),
                  Point([0.0, 2.0, 0.0]),
                  Point([0.0, 2.0, 2.0]),
                  Point([0.0, 0.0, 2.0])]
        self.assertEqual(compute_polygon_area(points), 4.0)

        # Test 3: triangle
        points = [Point([0.0, 0.0, 0.0]),
                  Point([0.0, 2.0, 0.0]),
                  Point([2.0, 0.0, 0.0])
                  ]
        self.assertEqual(compute_polygon_area(points), 2.0)
        # Test 4: real case
        points = [Point([2.5, 5.0, -0.5]),
                  Point([2.89216, 5.0, -1.0]),
                  Point([2.89216, 5.0, -0.820348]),
                  ]

    def test_compute_area_unordered(self):
        # Test 1
        points = [Point([0.0, 1.0, 0.0]),
                  Point([0.0, 1.0, 2.0]),
                  Point([0.0, 0.0, 0.0]),
                  Point([0.0, 0.0, 2.0])]
        self.assertEqual(compute_polygon_area(points), 2.0)

        # Test 2
        points = [Point([0.0, 1.0, 0.0]),
                  Point([2.0, 1.0, 0.0]),
                  Point([2.0, 1.0, 2.0]),
                  Point([0.0, 1.0, 2.0])]
        self.assertEqual(compute_polygon_area(points), 4.0)

    def test_filter_unique_points(self):
        point_test = [Point([0.0, 0.0, 0.0]), Point([1.0, 0.0, 0.0]), Point([0.0, 0.0, 0.0]), Point([0.0, 0.0, 0.0]), Point([1.0, 0.0, 0.0])]
        unique_points = filter_unique_points(point_test)
        self.assertEqual(len(unique_points), 2)
        point_test = [Point([2.0, 2.0001, 0.0]), Point([2.0, 2.0000001, 0.0]), Point([2.0, 2.00000001, 0.0]), Point([3.0, 1.0, 1.0])]
        unique_points = filter_unique_points(point_test)
        self.assertEqual(len(unique_points), 2)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

setup(
    name='pydelling',  # How you named your package folder (MyLib)
    packages=find_packages(),  # Chose the same as "name"
    include_package_data=True,
    version='1.7.10',  # Start with a small number and increase it with every change you make
    license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='Package to pre-process and help on mathematical modelling',  # Give a short description about your library
    author='Aitor Iraola Galarza',  # Type in your name
    author_email='aitirga@gmail.com',  # Type in your E-Mail
    url='https://github.com/aitirga/pydelling',  # Provide either the link to your github or to your website
    download_url='https://github.com/aitirga/pydelling/archive/refs/tags/v_1.7.10.zip',  # I explain this later on
    keywords=['PFLOTRAN', 'Preprocessing', 'python', 'modelling', 'postprocessing', 'math'],  # Keywords that define your package best
    install_requires=[  # I get to this in a second
        'ofpp',
        'numpy',
        'pandas',
        'h5py',
        'vtk',
        'matplotlib',
        'scipy',
        'python-box',
        'pyyaml',
        'setuptools',
        'xlrd',
        'openpyxl',
        'tqdm',
        'seaborn',
        'natsort',
        'sklearn',
        'meshio',
        'plotly',
        'shapely',
        'tabulate',
        'mkdocs',
        'mkdocs-material',
        'mkdocstrings',
        'streamlit',
        'dill',
        'trimesh',
        'mpmath',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3',  # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



import os

import h5py
import matplotlib.pyplot as plt
import numpy as np

from .BaseWriter import BaseWriter


class HDF5RasterWriter(BaseWriter):
    def __init__(self, filename, data=None, region_name="region", times=0.0, attributes={}, **kwargs):
        super().__init__(self, data=data, **kwargs)
        if self.info["interpolation"]["type"] == "regular_mesh":
            if len(self.data.shape) == 2:
                if self.data.shape[1] == 3:
                    self.data = self.centroid_transform_to_mesh()
                else:
                    self.data = self.transform_flatten_to_regular_mesh(data=self.data)
                    self.data = np.array(self.data)
                    plt.imshow(self.data[0, :, :])
                    plt.show()
                    self.data = np.swapaxes(np.array(self.data), 0, 1)
                    self.data = np.swapaxes(self.data, 1, 2)
            elif len(self.data.shape) == 3:
                _data = []
                for layer in self.data:
                    _data.append(np.array(self._centroid_transform_to_mesh(layer)))
                self.data = np.array(_data)
                self.data = np.swapaxes(np.array(_data), 0, 1)
                self.data = np.swapaxes(self.data, 1, 2)
        if self.data is not None:
            self.data_loaded = True
        self.region_name = region_name
        self.times = times
        self.attributes = attributes

    def transform_flatten_to_regular_mesh(self, data):
        assert len(self.data[1].shape) == 1, "A flatten data array needs to be given"
        aux_array = []
        if len(data.shape) == 2:
            for case in data:
                aux_array.append(np.reshape(case, (self.info["interpolation"]["n_y"], self.info["interpolation"]["n_x"])).T)
        return np.array(aux_array)

    def centroid_transform_to_mesh(self):
        assert len(self.data.shape) >= 2 and self.data.shape[1] == 3
        _data = self.data[:, 2]
        _data = np.reshape(_data, (self.info["interpolation"]["n_x"], self.info["interpolation"]["n_y"]))
        return _data

    def _centroid_transform_to_mesh(self, data):
        assert len(data.shape) >= 2 and data.shape[1] == 3
        _data = data[:, 2]
        _data = np.reshape(_data, (self.info["interpolation"]["n_x"], self.info["interpolation"]["n_y"]))
        return _data

    def add_default_attributes(self, hdf5_group: h5py.Dataset):
        dilatation_factor = self.info['interpolation']['dilatation_factor']
        l_x = np.abs(self.info["interpolation"]["x_max"] - self.info["interpolation"]["x_min"])
        l_x_dilatated = np.abs(self.info["interpolation"]["x_max"] - self.info["interpolation"]["x_min"]) * dilatation_factor
        delta_x = (l_x_dilatated - l_x) / 2.0
        dx = self.info["interpolation"]["d_x"] * dilatation_factor
        l_y = np.abs(self.info["interpolation"]["y_max"] - self.info["interpolation"]["y_min"])
        l_y_dilatated = np.abs(self.info["interpolation"]["y_max"] - self.info["interpolation"]["y_min"]) * dilatation_factor
        delta_y = (l_y_dilatated - l_y) / 2.0
        dy = self.info["interpolation"]["d_y"] * dilatation_factor

        hdf5_group.attrs.create('Dimension', "XY", dtype="S3")
        hdf5_group.attrs["Discretization"] = [dx, dy]
        hdf5_group.attrs["Origin"] = [self.info["interpolation"]["x_min"] - delta_x, self.info["interpolation"]["y_min"] - delta_y]
        hdf5_group.attrs["Interpolation_Method"] = "STEP"

    def run(self, filename=None):
        if filename is not None:
            self.filename = filename

        if self.check_data():
            if not os.path.exists(self.filename):
                h5temp = h5py.File(self.filename, "w")
                h5temp.close()
            with h5py.File(self.filename, "r+") as h5temp:
                try:
                    temp_group = h5temp.create_group(name=self.region_name)
                except ValueError as e:
                    print(f"ERROR writing HDF5 file: {e}")
                    print(f"INFO: Possible solution: use remove_output_file(filename=\"\")")
                    exit(1)
                temp_group.create_dataset("Times", data=self.times)
                temp_group.create_dataset("Data", data=self.data)
                # Adds default attributes to the group
                self.add_default_attributes(temp_group)
                # Extends the default attributes to the ones defined by the user
                if self.attributes is not {}:
                    for attribute in self.attributes:
                        temp_group.attrs[attribute] = self.attributes[attribute]
        else:
            print("Couldn't find data to dump!")



import unittest
import numpy.testing as nptest
import numpy as np
from pydelling.utils.geometry import Line
from pydelling.preprocessing import DfnPreprocessor, MeshPreprocessor
from pydelling.preprocessing.dfn_preprocessor import DfnUpscaler
from pydelling.utils import test_data_path

class IntersectionCase(unittest.TestCase):
    def test_full_intersection(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.1, z=0.0, dip=0, dip_dir=0, size=1.2
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])
        solution = [
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, 0.0, 0.0]),
            np.array([0.0, 0.5, 0.0]),
        ]
        nptest.assert_array_almost_equal(intersected_points, solution)

    def test_full_intersection_rotated(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.1, z=0.0, dip=20, dip_dir=50, size=1.2
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])
        solution = [
            np.array([0.62220399, 0.0, -0.12220399]),
            np.array([0.0, 0.0, 5.12772932e-02]),
            np.array([0.0, 5.85765892e-01, -8.57658923e-02]),
        ]
        nptest.assert_array_almost_equal(intersected_points, solution)

    def test_partial_intersection_1(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.1, z=0.5, dip=90, dip_dir=90, size=1.5
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])

        solution = [
            np.array([0.1,0.,-0.25]),
            np.array([0.1 ,  0.65 ,-0.25]),
            np.array([0.1 ,0. , 0.4]),
        ]
        # dfn_preprocessor.to_vtk('test_dfn.vtk')
        # mesh_preprocessor.to_vtk('test_element.vtk')
        # intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
        #     dfn_preprocessor[0], export_all_points=False)
        # print(intersected_points)
        # with open('computed_intersections.csv', 'w') as f:
        #     import csv
        #     writer = csv.writer(f)
        #     writer.writerows(intersected_points)

        nptest.assert_array_almost_equal(intersected_points, solution)

    def test_partial_intersection_2(self):

        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.1, z=0.5, dip=75, dip_dir=85, size=1.5
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])

        solution = [
            np.array([0.30360464 , 0.      ,   -0.22444437]),
            np.array([0.26325587 , 0.4611885 , -0.22444437]),
            np.array([0.14876171 ,0.   ,      0.35123829]),
        ]


        nptest.assert_array_almost_equal(intersected_points, solution)


    def test_partial_intersection_3(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.1, z=0.3, dip=75, dip_dir=65, size=1.3
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])

        solution = [
            np.array([0.33225466 , 0.   ,      -0.32785179]),
            np.array([0.   ,       0.71252241, -0.32785179]),
            np.array([0.12422918 ,0.   ,      0.37577082]),
            np.array([0.    ,      0.51272544 ,-0.01272544]),
        ]

        # dfn_preprocessor.to_vtk('test_dfn.vtk')
        # mesh_preprocessor.to_vtk('test_element.vtk')
        # intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
        #     dfn_preprocessor[0], export_all_points=False)
        # print(intersected_points)
        # with open('computed_intersections.csv', 'w') as f:
        #     import csv
        #     writer = csv.writer(f)
        #     writer.writerows(intersected_points)

        nptest.assert_array_almost_equal(intersected_points, solution)


    def test_partial_intersection_one_corner_inside(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.4, z=0.2, dip=75, dip_dir=65, size=0.7
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])

        solution = [
            np.array([0.16664204, 0.471432, -0.13807404]),
            np.array([0.25949777, 0.08819276, 0.15230947]),
            np.array([0.33001579, 0.12107585, -0.13807404]),
        ]
        # Sort arrays before comparing
        intersected_points = np.sort(intersected_points, axis=0)
        solution = np.sort(solution, axis=0)
        nptest.assert_array_almost_equal(intersected_points, solution)


    def test_no_intersection_fracture_inside(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.2, y=0.2, z=0.0, dip=60, dip_dir=45, size=0.2
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])

        solution = [
            np.array([0.30606602, 0.16464466, -0.08660254]),
            np.array([0.16464466, 0.30606602, -0.08660254]),
            np.array([0.09393398, 0.23535534, 0.08660254]),
            np.array([0.23535534, 0.09393398, 0.08660254]),
        ]
        nptest.assert_array_almost_equal(intersected_points, solution)

    def test_intersect_hexahedra_full(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.0, y=0.0, z=0.0, dip=1, dip_dir=0, size=2.0
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_hexahedra(
            node_ids=np.array([0, 1, 2, 3, 4, 5, 6, 7]),
            node_coords=[
                np.array([-0.5, -0.5, -0.5]),
                np.array([0.5, -0.5, -0.5]),
                np.array([0.5, 0.5, -0.5]),
                np.array([-0.5, 0.5, -0.5]),
                np.array([-0.5, -0.5, 0.5]),
                np.array([0.5, -0.5, 0.5]),
                np.array([0.5, 0.5, 0.5]),
                np.array([-0.5, 0.5, 0.5])
            ],
        )

        intersections = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])
        solution = np.array([
            np.array([-0.5, -0.5, 0.00872753]),
            np.array([-0.5, 0.5, -0.00872753]),
            np.array([0.5, -0.5, 0.00872753]),
            np.array([0.5, 0.5, -0.00872753]),
        ])
        # Sort intersections to make them comparable
        intersections = np.sort(intersections, axis=0)
        solution = np.sort(solution, axis=0)

        nptest.assert_array_almost_equal(intersections, solution)

    def test_problematic_fracture_1(self):
        import pickle
        dfn_preprocessor = DfnPreprocessor()
        mesh = MeshPreprocessor()
        read_fracture = test_data_path() / "problematic_fractures/issue_fracture_1.pkl"
        with open(read_fracture, "rb") as f:
            fracture = pickle.load(f)
        read_element = test_data_path() / "problematic_fractures/issue_element_1.pkl"
        with open(read_element, "rb") as f:
            element = pickle.load(f)

        dfn_preprocessor.dfn.append(fracture)
        dfn_preprocessor.to_vtk('test_dfn.vtk')
        element.to_obj('test_element.obj')
        intersections = element.intersect_with_fracture(dfn_preprocessor[0], export_all_points=True)
        with open('computed_intersections.csv', 'w') as f:
            import csv
            writer = csv.writer(f)
            writer.writerows(intersections)
        self.assertEqual(len(intersections), 3)

    def test_problematic_fracture_2(self):
        import pickle
        dfn_preprocessor = DfnPreprocessor()
        mesh = MeshPreprocessor()
        read_fracture = test_data_path() / "problematic_fractures/issue_fracture_2.pkl"
        with open(read_fracture, "rb") as f:
            fracture = pickle.load(f)
        read_element = test_data_path() / "problematic_fractures/issue_element_2.pkl"
        with open(read_element, "rb") as f:
            element = pickle.load(f)

        dfn_preprocessor.dfn.append(fracture)
        mesh.elements.append(element)
        intersections = mesh.elements[0].intersect_with_fracture(dfn_preprocessor[0])
        self.assertEqual(len(intersections), 4)






if __name__ == "__main__":
    unittest.main()



import logging
from pathlib import Path
from typing import List

import meshio
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from tabulate import tabulate
from tqdm import tqdm

from .Fault import Fault
from .Fracture import Fracture

logger = logging.getLogger(__name__)


class DfnPreprocessor(object):
    dfn: List[Fracture] = []
    faults: List[Fault] = []

    def __getitem__(self, item):
        return self.dfn[item]

    def __init__(self):
        self.clean_dfn()

    def clean_dfn(self):
        """Removes dfn object"""
        self.dfn = []
        self.faults = []


    def load_fractures(self, pd_df: pd,
                       dip='dip',
                       dip_dir='dip-direction',
                       x='position-x',
                       y='position-y',
                       z='position-z',
                       size='size',
                       aperture=None,
                       aperture_constant=None,
                       transmissivity_constant=None,
                       storativity_constant=None
                       ):
        """
        Loads the fractures from a pandas dataframe to the dfn object.
        Args:
            pd_df: pandas dataframe containing the fractures

        Returns:

        """
        for index, row in tqdm(pd_df.iterrows()):
            self.add_fracture(
                dip=row[dip],
                dip_dir=row[dip_dir],
                x=row[x],
                y=row[y],
                z=row[z],
                size=row[size],
                aperture_constant=aperture_constant,
                aperture=aperture,
                transmissivity_constant=transmissivity_constant,
                storativity_constant=storativity_constant
            )

    def load_fractures_from_polygons_and_apertures(self,
                       polygons,
                       apertures=None,
                       radii=None,
                       aperture_constant=None,
                       transmissivity_constant=None,
                       storativity_constant=None):
        logger.info('Loading fractures from polygons and apertures')
        for idx, polygon in tqdm(enumerate(polygons), desc='Loading fractures into the DFN', total=len(polygons)):
            self.add_fracture(
                polygon=polygon,
                aperture=apertures[idx] if apertures is not None else None,
                size=radii[idx] * 2 if radii is not None else None,
                aperture_constant=aperture_constant,
                transmissivity_constant=transmissivity_constant,
                storativity_constant=storativity_constant,
            )


    def add_fracture(self,
                     x=None,
                     y=None,
                     z=None,
                     dip=None,
                     dip_dir=None,
                     size=None,
                     aperture=None,
                     aperture_constant=1E-3,
                     transmissivity_constant=None,
                     storativity_constant=None,
                     polygon=None,
                     ):
        """Add individual fracture to the dfn object.
        """
        self.dfn.append(Fracture(
            dip=dip,
            dip_dir=dip_dir,
            x=x,
            y=y,
            z=z,
            size=size,
            aperture=aperture,
            aperture_constant=aperture_constant,
            transmissivity_constant=transmissivity_constant,
            storativity_constant=storativity_constant,
            polygon=polygon,
        ))

    def add_fault(self, filename=None,
                  mesh=None,
                  aperture=None):
        """Adds a fault to the dfn object."""
        if aperture is None:
            logger.warning(f'No aperture specified for fault {filename}')
        if isinstance(filename, Fault):
            self.faults.append(filename)
        elif isinstance(filename, str) or isinstance(filename, Path):
            self.faults.append(Fault(filename=filename, mesh=mesh, aperture=aperture))
        else:
            logger.error('Fault filename must be a string or Fault object')
            raise TypeError('Fault filename must be a string or Fault object')

    def summary(self):
        """Prints a summary of the dfn object."""
        print(tabulate(
            [
                ['Number of fractures', len(self.dfn)],
                ['Max size', f"{self.max_size:1.2f} m"],
                ['Min size', f"{self.min_size:1.2f} m"]
            ],
            headers=['Parameter', 'Value'],
            tablefmt='grid',
            numalign='center',
        ))

    def __len__(self):
        return len(self.dfn)

    def visualize_dfn(self, add_centroid=True, fracture_color='blue', size_color=False):
        """Visualizes the dfn object."""
        self.fig = self.generate_dfn_plotly(add_centroid=add_centroid, fracture_color=fracture_color, size_color=size_color)
        self.fig.show()

    def export_dfn_image(self, filename='dfn.png', add_centroid=True, fracture_color='blue', *args, **kwargs, ):
        logger.info(f'Exporting dfn image to {filename}')
        self.fig = self.generate_dfn_plotly(add_centroid=add_centroid, fracture_color=fracture_color)
        self.fig.write_image(filename, *args, **kwargs)

    def to_obj(self, filename='dfn.obj', method='v1'):
        '''Exports the dfn object to stl format.'''
        logger.info(f'Exporting dfn object to {filename}')
        obj_file = open(filename, 'w')
        obj_file.write('# Created by pydelling\n')
        obj_file.write('o dfn\n')
        global_id = 1
        for fracture in tqdm(self.dfn):
            obj_file.write(fracture.to_obj(global_id=global_id, method=method))
            global_id += fracture.n_side_points
        for fault in self.faults:
            fault_obj = fault.to_obj(global_id=global_id)
            obj_file.write(fault_obj)
            global_id += fault.num_points

    def to_vtk(self, filename='dfn.vtk', method='v1'):
        from pathlib import Path
        self.to_obj('buffer.obj', method=method)
        meshio_mesh: meshio.Mesh = meshio.read('buffer.obj')
        meshio_mesh.cell_data = {'aperture': self.apertures}
        meshio.write(filename, meshio_mesh, file_format='vtk')
        Path('buffer.obj').unlink()




    def to_dfnworks(self, filename='dfn.dat', method='v1'):
        '''Exports the dfn object to dfnworks format.'''
        logger.info(f'Exporting dfn object to {filename}')
        dfn_file = open(filename, 'w')
        n_total_fractures = len(self.dfn)
        dfn_file.write(f'nPolygons: {n_total_fractures}\n')
        for fracture in tqdm(self.dfn):
            side_points = fracture.get_side_points(method=method)
            dfn_file.write(f'{len(side_points)} ')
            for point in side_points:
                dfn_file.write(f'{{{point[0]},{point[1]},{point[2]}}}')
            dfn_file.write('\n')

    def shift(self, x_shift=0, y_shift=0, z_shift=0):
        """Shifts the dfn object."""
        logger.info(f'Shifting dfn object by {x_shift}, {y_shift}, {z_shift}')
        for fracture in self.dfn:
            fracture.shift(x_shift, y_shift, z_shift)


    def generate_dfn_plotly(self, add_centroid=False, size_color=False, fracture_color='blue'):
        ''' Generates a plotly figure of the dfn object.

        Returns: A plotly figure

        '''
        logger.info('Generating plotly figure')
        fig = go.Figure()
        for fracture in tqdm(self.dfn):
            fracture_sides = fracture.get_side_points()
            if add_centroid:
                fig.add_trace(go.Scatter3d(
                    x=[fracture.x_centroid],
                    y=[fracture.y_centroid],
                    z=[fracture.z_centroid],
                    mode='markers',
                    marker=dict(
                        size=3.5,
                        color='black',
                        symbol='circle',
                        opacity=0.65
                    )
                ))
            # Add color depending on the fracture size
            if size_color:
                A = 255 / (self.max_size - self.min_size)
                B = 255 - A * self.max_size
                value = int(A * fracture.size + B)
                color = f'rgb({value}, 0, 0)'
            else:
                color = fracture_color
            fig.add_trace(go.Mesh3d(
                x=fracture_sides[:, 0],
                y=fracture_sides[:, 1],
                z=fracture_sides[:, 2],
                color=color,
                opacity=0.75,
            ))

        return fig

    @property
    def max_size(self):
        """Returns the maximum size of the dfn object."""
        return max([fracture.size for fracture in self.dfn])

    @property
    def min_size(self):
        """Returns the minimum size of the dfn object."""
        return min([fracture.size for fracture in self.dfn])

    def plot_radii_histogram(self, filename='radii_histogram.png'):
        """Plots the radii histogram."""
        logger.info(f'Plotting radii histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([(fracture.size / 2) for fracture in self.dfn], bins=100)
        return fig, ax

    def plot_aperture_histogram(self, filename='aperture_histogram.png'):
        """Plots the aperture histogram."""
        logger.info(f'Plotting aperture histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([fracture.aperture for fracture in self.dfn], bins=100)
        return fig, ax

    def plot_transmissivity_histogram(self, filename='transmissivity_histogram.png'):
        """Plots the transmissivity histogram."""
        logger.info(f'Plotting aperture histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([fracture.transmissivity for fracture in self.dfn], bins=100)
        return fig, ax

    def plot_hkx_histogram(self, filename='hkx_histogram.png'):
        """Plots the x-hydraulic conductivity histogram."""
        logger.info(f'Plotting hk_x histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([(fracture.transmissivity / fracture.aperture) for fracture in self.dfn], bins=100)
        return fig, ax


    def plot_storativity_histogram(self, filename='storativity_histogram.png'):
        """Plots the storativity histogram."""
        logger.info(f'Plotting aperture histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([fracture.storativity for fracture in self.dfn], bins=100)
        return fig, ax

    @property
    def apertures(self) -> np.ndarray:
        return np.array([fracture.aperture for fracture in self.dfn])

    def __add__(self, other):
        """Adds two dfn objects."""
        if not isinstance(other, DfnPreprocessor):
            raise TypeError(f'{other} is not a DfnPreprocessor object')

        logger.info('Adding dfn objects')
        new_dfn = DfnPreprocessor()
        new_dfn.dfn = self.dfn + other.dfn
        new_dfn.faults = self.faults + other.faults
        return new_dfn

    @property
    def min_x(self):
        return min([fracture.x_centroid for fracture in self.dfn])

    @property
    def max_x(self):
        return max([fracture.x_centroid for fracture in self.dfn])

    @property
    def min_y(self):
        return min([fracture.y_centroid for fracture in self.dfn])

    @property
    def max_y(self):
        return max([fracture.y_centroid for fracture in self.dfn])

    @property
    def min_z(self):
        return min([fracture.z_centroid for fracture in self.dfn])

    @property
    def max_z(self):
        return max([fracture.z_centroid for fracture in self.dfn])

    def get_json(self):
        """Returns a json representation of the dfn object."""
        export_dict = {}
        export_dict['dfn'] = [fracture.get_json() for fracture in self.dfn]
        export_dict['faults'] = [fault.get_json() for fault in self.faults]
        return export_dict

    def to_json(self, filename):
        """Writes the dfn object to a json file."""
        import json
        with open(filename, 'w') as f:
            json.dump(self.get_json(), f)

    @classmethod
    def from_json(cls, filename='dfn.json'):
        """Loads a dfn object from a json file."""
        import json
        with open(filename, 'r') as f:
            Fracture.local_id = 0  # Be careful with this
            Fault.local_id = 0  # Be careful with this
            dfn_dict = json.load(f)
            dfn_object = cls()
            dfn_object.dfn = [Fracture(**fracture) for fracture in dfn_dict['dfn']]
            dfn_object.faults = [Fault(**fault) for fault in dfn_dict['faults']]
            return dfn_object

    @classmethod
    def from_dict(cls, dict: dict):
        """Loads a dfn object from a dict."""
        Fracture.local_id = 0
        Fault.local_id = 0
        dfn_object = cls()
        dfn_object.dfn = [Fracture(**fracture) for fracture in dict['dfn']]
        dfn_object.faults = [Fault(**fault) for fault in dict['faults']]
        return dfn_object


    def __repr__(self):
        return f'DFN with {len(self.dfn)} fractures and {len(self.faults)} faults'

    def __str__(self):
        return self.__repr__()









import unittest
from pydelling.readers import CentroidReader
from pydelling.config import config
from pydelling.utils import test_data_path
import logging
# logging.getLogger().setLevel("CRITICAL")
from pathlib import Path


class CentroidReaderCase(unittest.TestCase):
    """
    This class tests the iGPReader[pydelling.readers.iGPReader.io.iGPReader.iGPReader] class implementation
    """
    def setUp(self) -> None:
        self.centroid_reader = CentroidReader(test_data_path() / "centroid_reader_data.csv", header=True, separator=",")

    def test_centroid_data(self):
        self.assertEqual(self.centroid_reader.get_data(as_dataframe=False)[0][0], 0.0)

    def test_write_data(self):
        write_path = Path().cwd() / "test.csv"
        self.centroid_reader.dump_to_csv(write_path)
        write_path.unlink()


if __name__ == '__main__':
    unittest.main()




import numpy.testing as nptest
import numpy as np
from pydelling.utils.geometry import Line, Plane, Vector, Point

if __name__ == '__main__':
    # plane_1 = Plane(point=[0, 0, 3], normal=Vector([0, 0, 1]))
    # plane_2 = Plane(point=[0, -10, 0], normal=Vector([0, -1, 0]))
    # print(plane_1.intersect(plane_2))
    # line_1 = Line(p1=[0, 0, 0], direction_vector=[-1, 0, 0])
    # line_2 = Line(p1=[0.5, 0.5, 0], direction_vector=[1, -1, 0])
    # print(line_1.intersect(line_2))

    # line_1 = Line(p1=[0, 0, 0], direction_vector=[-1, 0, 0])
    # line_2 = Line(p1=[-1, 0, 0], direction_vector=[0, 1, 0])
    # print(line_1.intersect(line_2))
    # test_point: Point = Point([0, 0, 0])
    test_vector: Vector = Vector(p1=[-1, 0, 0], p2=[1, 1, 1])

